/*
** Destop's player plugin
** Copyright (c) 2003 Destop - CZN
** w/ modifications by hcs
*/

#include <windows.h>
#include "dsp.h"
#include "wamain.h"

DSPFILE dspfile;

int get16bit(unsigned char* p)
{
	return (p[0] << 8) | p[1];
}

int get32bit(unsigned char* p)
{
	return (p[0] << 24) | (p[1] << 16) | (p[2] << 8) | p[3];
}

// standard devkit version
void get_dspheaderstd(DSPHEAD *dsp,unsigned char *buf)
{
	int i;
	dsp->num_samples=get32bit(buf);
	dsp->num_adpcm_nibbles=get32bit(buf+4);
	dsp->sample_rate=get32bit(buf+8);
	dsp->loop_flag=get16bit(buf+0xC);
	dsp->format=get16bit(buf+0xE);
	dsp->sa=get32bit(buf+0x10);
	dsp->ea=get32bit(buf+0x14);
	dsp->ca=get32bit(buf+0x18);
	DisplayError("get_dspheaderstd:\nnum_samples=%li\nnum_adpcm_nibbles=%li\nsample_rate=%li\n"
		"loop_flag=%04x\nformat=%04x\nsa=%08x\nea=%08x\nca=%08x",dsp->num_samples,dsp->num_adpcm_nibbles,
		dsp->sample_rate,dsp->loop_flag,dsp->format,dsp->sa,dsp->ea,dsp->ca);
	for (i=0;i<16;i++)
	dsp->coef[i]=get16bit(buf+0x1C+i*2);
	dsp->gain=get16bit(buf+0x3C);
	dsp->ps=get16bit(buf+0x3E);
	dsp->yn1=get16bit(buf+0x40);
	dsp->yn2=get16bit(buf+0x42);
	dsp->lps=get16bit(buf+0x44);
	dsp->lyn1=get16bit(buf+0x46);
	dsp->lyn2=get16bit(buf+0x48);
}

// SF Assault version
void get_dspheadersfa(DSPHEAD *dsp,unsigned char *buf)
{
	int i;
	buf+=0x20; // for SF Assault
	dsp->num_samples=get32bit(buf);
	dsp->num_adpcm_nibbles=get32bit(buf+4);
	dsp->sample_rate=get32bit(buf+8);
	dsp->loop_flag=get16bit(buf+0xC);
	dsp->format=get16bit(buf+0xE);
	dsp->sa=get32bit(buf+0x10);
	dsp->ea=get32bit(buf+0x14);
	dsp->ca=get32bit(buf+0x18);
	DisplayError("get_dspheadersfa:\nnum_samples=%li\nnum_adpcm_nibbles=%li\nsample_rate=%li\n"
		"loop_flag=%04x\nformat=%04x\nsa=%08x\nea=%08x\nca=%08x",dsp->num_samples,dsp->num_adpcm_nibbles,
		dsp->sample_rate,dsp->loop_flag,dsp->format,dsp->sa,dsp->ea,dsp->ca);
	for (i=0;i<16;i++)
	dsp->coef[i]=get16bit(buf+0x1C+i*2);
	dsp->gain=get16bit(buf+0x3C);
	dsp->ps=get16bit(buf+0x3E);
	dsp->yn1=get16bit(buf+0x40);
	dsp->yn2=get16bit(buf+0x42);
	dsp->lps=get16bit(buf+0x44);
	dsp->lyn1=get16bit(buf+0x46);
	dsp->lyn2=get16bit(buf+0x48);
}

// Metroid Prime 2 version
void get_dspheadermp2(DSPHEAD *dsp,unsigned char *buf)
{
	int i;
	dsp->num_samples=get32bit(buf+0x8);
	dsp->num_adpcm_nibbles=get32bit(buf+0x10);
	dsp->sample_rate=get32bit(buf+0x0c);
	dsp->loop_flag=get16bit(buf+0x14);
	dsp->format=get16bit(buf+0xE);
	dsp->sa=get32bit(buf+0x18);
	dsp->ea=get32bit(buf+0x1c);
	DisplayError("get_dspheadermp2:\nnum_samples=%li\nnum_adpcm_nibbles=%li\nsample_rate=%li\n"
		"loop_flag=%04x\nformat=%04x\nsa=%08x\nea=%08x\nca=%08x",dsp->num_samples,dsp->num_adpcm_nibbles,
		dsp->sample_rate,dsp->loop_flag,dsp->format,dsp->sa,dsp->ea,dsp->ca);
	for (i=0;i<16;i++)
	dsp->coef[i]=get16bit(buf+0x20+i*2);
	dsp->yn1=dsp->yn2=dsp->lyn1=dsp->lyn2=0;
}

// Metroid Prime 2 version (second channel)
void get_dspheadermp22(DSPHEAD *dsp,unsigned char *buf)
{
	int i;
	dsp->num_samples=get32bit(buf+0x8);
	dsp->num_adpcm_nibbles=get32bit(buf+0x10);
	dsp->sample_rate=get32bit(buf+0x0c);
	dsp->loop_flag=get16bit(buf+0x14);
	dsp->format=get16bit(buf+0xE);
	dsp->sa=get32bit(buf+0x18);
	dsp->ea=get32bit(buf+0x1c);
	DisplayError("get_dspheadermp22:\nnum_samples=%li\nnum_adpcm_nibbles=%li\nsample_rate=%li\n"
		"loop_flag=%04x\nformat=%04x\nsa=%08x\nea=%08x\nca=%08x",dsp->num_samples,dsp->num_adpcm_nibbles,
		dsp->sample_rate,dsp->loop_flag,dsp->format,dsp->sa,dsp->ea,dsp->ca);
	for (i=0;i<16;i++)
	dsp->coef[i]=get16bit(buf+0x40+i*2);
	dsp->yn1=dsp->yn2=dsp->lyn1=dsp->lyn2=0;
}

// SSB:M HALPST version
void get_dspheaderhalp(DSPHEAD *dsp,unsigned char *buf)
{
	int i;
	dsp->num_samples=get32bit(buf+0x18); // I'm using the same as the loop endpoint...
	dsp->num_adpcm_nibbles=get32bit(buf+0x18)*2; // ditto
	dsp->sample_rate=get32bit(buf+0x08);
	dsp->loop_flag=get16bit(buf+0x10);
	dsp->format=get16bit(buf+0xE);
	dsp->sa=get32bit(buf+0x14);
	dsp->ea=get32bit(buf+0x18);
	DisplayError("get_dspheaderhalp:\nnum_samples=%li\nnum_adpcm_nibbles=%li\nsample_rate=%li\n"
		"loop_flag=%04x\nformat=%04x\nsa=%08x\nea=%08x\nca=%08x",dsp->num_samples,dsp->num_adpcm_nibbles,
		dsp->sample_rate,dsp->loop_flag,dsp->format,dsp->sa,dsp->ea,dsp->ca);
	for (i=0;i<16;i++)
	dsp->coef[i]=get16bit(buf+0x20+i*2);
	dsp->yn1=dsp->yn2=dsp->lyn1=dsp->lyn2=0;
}

// SSB:M HALPST version (second channel)
void get_dspheaderhalp2(DSPHEAD *dsp,unsigned char *buf)
{
	int i;
	dsp->num_samples=get32bit(buf+0x50); // I'm using the same as the loop endpoint...
	dsp->num_adpcm_nibbles=get32bit(buf+0x50)*2; // ditto
	dsp->sample_rate=get32bit(buf+0x08);
	dsp->loop_flag=get16bit(buf+0x48);
	dsp->format=get16bit(buf+0x4a);
	dsp->sa=get32bit(buf+0x4c);
	dsp->ea=get32bit(buf+0x50);
	DisplayError("get_dspheaderhalp2:\nnum_samples=%li\nnum_adpcm_nibbles=%li\nsample_rate=%li\n"
		"loop_flag=%04x\nformat=%04x\nsa=%08x\nea=%08x\nca=%08x",dsp->num_samples,dsp->num_adpcm_nibbles,
		dsp->sample_rate,dsp->loop_flag,dsp->format,dsp->sa,dsp->ea,dsp->ca);
	for (i=0;i<16;i++)
	dsp->coef[i]=get16bit(buf+0x58+i*2);
	dsp->yn1=dsp->yn2=dsp->lyn1=dsp->lyn2=0;
}

// 
void InitDSPFILE(HANDLE infile, DSPFILE * dsp) {
	unsigned char readbuf[0x100];
	int l;

	dsp->ch[0].infile=dsp->ch[1].infile=infile;

	SetFilePointer(dsp->ch[0].infile,0,0,FILE_BEGIN);

	ReadFile(dsp->ch[0].infile, &readbuf, 0x100, &l, NULL);
	
	if (readbuf[0]=='R' && readbuf[1]=='S' && readbuf[2]==0x00 && readbuf[3]==0x03) {
		// Metroid Prime 2 "RS03" (looping still not working right)
		if (get16bit(readbuf+6)==2) { // channel count
			get_dspheadermp2(&dsp->ch[0].header,readbuf);
			get_dspheadermp22(&dsp->ch[1].header,readbuf);

			dsp->ch[0].header.sa/=2; //dsp->ch[0].header.ea-dsp->ch[0].header.sa*2;
			dsp->ch[1].header.sa/=2; //dsp->ch[1].header.ea-dsp->ch[1].header.sa*2;

			dsp->NCH=2;
			dsp->ch[0].interleave=dsp->ch[1].interleave=0x8f00;
		
			dsp->ch[0].chanstart=0x60;
			dsp->ch[1].chanstart=0x8f60;
		
			dsp->ch[0].type=dsp->ch[1].type=type_mp2;
			dsp->ch[0].bps=dsp->ch[1].bps=8;
		} else {
			// mono variant, haven't seen any of these loop yet
			get_dspheadermp2(&dsp->ch[0].header,readbuf);
			
			dsp->ch[0].header.sa/=2;

			dsp->NCH=1;
			dsp->ch[0].interleave=0;

			dsp->ch[0].chanstart=0x60;

			dsp->ch[0].type=type_mp2;
			dsp->ch[0].bps=dsp->ch[1].bps=4;
		}
	} else if (readbuf[0]=='C' && readbuf[1]=='s' && readbuf[2]=='t' && readbuf[3]=='r') {
		// Star Fox Assault "Cstr"
		get_dspheadersfa(&dsp->ch[0].header,readbuf);
		get_dspheadersfa(&dsp->ch[1].header,readbuf+0x60);

		// weird but needed
		dsp->ch[0].header.sa*=2;

		// second header has odd values for sa and ea
		dsp->ch[1].header.ea=dsp->ch[0].header.ea;
		dsp->ch[1].header.sa=dsp->ch[0].header.sa;

		dsp->NCH=2;
		dsp->ch[0].interleave=dsp->ch[1].interleave=0x800;

		dsp->ch[0].chanstart=0xe0;
		dsp->ch[1].chanstart=0x8e0;

		dsp->ch[0].type=dsp->ch[1].type=type_sfass;
		dsp->ch[0].bps=dsp->ch[1].bps=8;
	} else if (readbuf[0]==0x02 && readbuf[1]==0 && readbuf[2]==0x7d && readbuf[3]==0) {
		// Paper Mario 2 "STM"
		get_dspheaderstd(&dsp->ch[0].header,readbuf+0x40);
		get_dspheaderstd(&dsp->ch[1].header,readbuf+0xa0);
		
		dsp->NCH=2;
		dsp->ch[0].interleave=dsp->ch[1].interleave=0;

		dsp->ch[0].chanstart=0x100;
		dsp->ch[1].chanstart=0x100+get32bit(readbuf+8); // the offset of chan2 is in several places...

		dsp->ch[0].type=dsp->ch[1].type=type_pm2;
		dsp->ch[0].bps=dsp->ch[1].bps=4;
	} else if (readbuf[0]==' ' && readbuf[1]=='H' && readbuf[2]=='A' && readbuf[3]=='L' && readbuf[4]=='P') {
		// Super Smash Bros. Melee "HALPST"
		get_dspheaderhalp(&dsp->ch[0].header,readbuf);
		get_dspheaderhalp2(&dsp->ch[1].header,readbuf);

		dsp->NCH=2;
		dsp->ch[0].interleave=dsp->ch[1].interleave=0x8000;
		
		dsp->ch[0].chanstart=0xa0;
		dsp->ch[1].chanstart=0x80a0;
		
		dsp->ch[0].type=dsp->ch[1].type=type_halp;
		dsp->ch[0].bps=dsp->ch[1].bps=8;
	} else {
		// assume standard devkit
		get_dspheaderstd(&dsp->ch[0].header,readbuf);
		get_dspheaderstd(&dsp->ch[1].header,readbuf+0x60); // header for channel 2 (duh!)

		// if a valid second header (agrees with first)
		if (dsp->ch[0].header.num_adpcm_nibbles==dsp->ch[1].header.num_adpcm_nibbles &&
			dsp->ch[0].header.num_samples==dsp->ch[1].header.num_samples) {

			// stereo
			dsp->NCH=2;
			dsp->ch[0].interleave=dsp->ch[1].interleave=0x8000;

			dsp->ch[0].chanstart=0xc0;
			dsp->ch[1].chanstart=0x80c0;

			dsp->ch[0].type=dsp->ch[1].type=type_std;
			dsp->ch[0].bps=dsp->ch[1].bps=8;
		} else {
			// mono
			dsp->NCH=1;
			dsp->ch[0].interleave=0;

			// a Metroid peculiarity?
			if (dsp->ch[0].header.sa&0xf != 0x02 && dsp->ch[0].header.sa&0xf != 0x0a) {
				dsp->ch[0].header.sa/=2;
				DisplayError("MP2demo");
			}

			dsp->ch[0].chanstart=0x60;

			dsp->ch[0].type=type_std;

			dsp->ch[0].bps=4;
		}
	}

	dsp->ch[0].offs=dsp->ch[0].chanstart;
	dsp->ch[1].offs=dsp->ch[1].chanstart;

	if (!dsp->ch[0].header.loop_flag) dsp->nrsamples = dsp->ch[0].header.num_samples;
	else if (dsp->ch[0].interleave)
		dsp->nrsamples=(dsp->ch[0].header.sa+looptimes*(dsp->ch[0].header.ea-dsp->ch[0].header.sa))*14/(8*8/dsp->ch[0].bps)/dsp->NCH+masterfadelength*dsp->ch[0].header.sample_rate;
	else
		dsp->nrsamples=(dsp->ch[0].header.sa+looptimes*(dsp->ch[0].header.ea-dsp->ch[0].header.sa))*14/(8*8/dsp->ch[0].bps)+masterfadelength*dsp->ch[0].header.sample_rate;

	dsp->file_length=GetFileSize(dsp->ch[0].infile,NULL);

	dsp->ch[0].hist1=dsp->ch[0].header.yn1;
	dsp->ch[0].hist2=dsp->ch[0].header.yn2;
	dsp->ch[1].hist1=dsp->ch[1].header.yn1;
	dsp->ch[1].hist2=dsp->ch[1].header.yn2;

	dsp->ch[0].readloc=dsp->ch[1].readloc=dsp->ch[0].writeloc=dsp->ch[1].writeloc=0;
}

// Note that for both of these I don't bother loading the loop context. Sounds great the
// way it is, why mess with a good thing?

// for noninterleaved files (mono, STM)
// also possibly for reading two mono Metroid Prime files simultaneously as stereo
void fillbuffer(DSPSTREAM * stream) {
	int i,l;
	short decodebuf[14];
	char ADPCMbuf[8];

	SetFilePointer(stream->infile,stream->offs,0,FILE_BEGIN);

	i=0;
	do {
		if (i==0) {
			ReadFile(stream->infile, ADPCMbuf, 8, &l, NULL);
			decodebuffer(ADPCMbuf,decodebuf,stream->header.coef,&stream->hist1,&stream->hist2);
			i=14;
			stream->offs+=8;
			
			// seems to work OK for everything but MP2 demo (but now that only has a slight loop issue)
			if (stream->header.loop_flag && (stream->offs-stream->chanstart)>=stream->header.ea*stream->bps/8)
				stream->offs=SetFilePointer(stream->infile,stream->chanstart+(stream->header.sa&(~0x7))*stream->bps/8,0,FILE_BEGIN);
		}
		stream->chanbuf[stream->writeloc++]=decodebuf[14-i];
		i--;
		if (stream->writeloc>=0x8000/8*14) stream->writeloc=0;
	} while (stream->writeloc != stream->readloc);
}

// interleaved files requires streams with knowledge of each other (for proper looping)
// will need another version of this for HALPST?
void fillbufferinterleave(DSPFILE * dsp) {
	int i,l;
	short decodebuf1[14];
	short decodebuf2[14];
	char ADPCMbuf1[8];
	char ADPCMbuf2[8];
	
	i=0;
	do {
		if (i==0) {
			SetFilePointer(dsp->ch[0].infile, dsp->ch[0].offs,0,FILE_BEGIN);
			ReadFile(dsp->ch[0].infile, ADPCMbuf1, 8, &l, NULL);

			SetFilePointer(dsp->ch[1].infile, dsp->ch[1].offs,0,FILE_BEGIN);
			ReadFile(dsp->ch[1].infile, ADPCMbuf2, 8, &l, NULL);
			
			// handle HALPST headers (alternative to interleave)
			if (dsp->ch[0].type==type_halp && !ADPCMbuf2[0]) {
				u32 size=get32bit(ADPCMbuf2);
				dsp->ch[0].offs=dsp->ch[1].offs+0x20;

				SetFilePointer(dsp->ch[1].infile, dsp->ch[1].offs+8,0,FILE_BEGIN);
				ReadFile(dsp->ch[1].infile, ADPCMbuf2, 8, &l, NULL);
				
				dsp->ch[1].offs=get32bit(ADPCMbuf2)-size/2;

				SetFilePointer(dsp->ch[1].infile, dsp->ch[1].offs,0,FILE_BEGIN);
				ReadFile(dsp->ch[1].infile, ADPCMbuf2, 8, &l, NULL);
			
				SetFilePointer(dsp->ch[0].infile, dsp->ch[0].offs,0,FILE_BEGIN);
				ReadFile(dsp->ch[0].infile, ADPCMbuf1, 8, &l, NULL);
			}
			decodebuffer(ADPCMbuf1,decodebuf1,dsp->ch[0].header.coef,&dsp->ch[0].hist1,&dsp->ch[0].hist2);
			decodebuffer(ADPCMbuf2,decodebuf2,dsp->ch[1].header.coef,&dsp->ch[1].hist1,&dsp->ch[1].hist2);

			i=14;
			dsp->ch[0].offs+=8;
			dsp->ch[1].offs+=8;

			// handle interleave
			if (dsp->ch[0].type!=type_halp) {
				if ((dsp->ch[0].offs-dsp->ch[0].chanstart)%dsp->ch[0].interleave==0)
					dsp->ch[0].offs+=dsp->ch[0].interleave;
				if ((dsp->ch[1].offs-dsp->ch[1].chanstart)%dsp->ch[1].interleave==0)
					dsp->ch[1].offs+=dsp->ch[1].interleave;
			}
						
			if (dsp->ch[0].header.loop_flag && (
				(dsp->ch[0].offs-dsp->ch[0].chanstart)>=dsp->ch[0].header.ea*dsp->ch[0].bps/8 ||
				(dsp->ch[1].offs-dsp->ch[0].chanstart)>=dsp->ch[1].header.ea*dsp->ch[1].bps/8 ) ) { // use if dsp->ch[0].chanstart on this line is intentional
				dsp->ch[0].offs=SetFilePointer(dsp->ch[0].infile,dsp->ch[0].chanstart+(dsp->ch[0].header.sa&(~7))*dsp->ch[0].bps/8,0,FILE_BEGIN);
				dsp->ch[1].offs=SetFilePointer(dsp->ch[1].infile,dsp->ch[1].chanstart+(dsp->ch[1].header.sa&(~7))*dsp->ch[1].bps/8,0,FILE_BEGIN);
				//DisplayError("loop");
			}
		}
		dsp->ch[0].chanbuf[dsp->ch[0].writeloc++]=decodebuf1[14-i];
		dsp->ch[1].chanbuf[dsp->ch[1].writeloc++]=decodebuf2[14-i];
		i--;
		if (dsp->ch[0].writeloc>=0x8000/8*14) dsp->ch[0].writeloc=0;
		if (dsp->ch[1].writeloc>=0x8000/8*14) dsp->ch[1].writeloc=0;
	} while (dsp->ch[0].writeloc != dsp->ch[0].readloc);
}

void fillbuffers(DSPFILE * dsp) {
	if (dsp->ch[0].interleave) {
		fillbufferinterleave(dsp);
	} else {
		fillbuffer(&dsp->ch[0]);
		if (dsp->NCH==2)
			fillbuffer(&dsp->ch[1]);
	}
}